@@ -38,6 +38,12 @@ using Mono.Addins.Description;
 
 namespace Mono.Addins.Setup
 {
+	/// <summary>
+	/// A command line add-in manager.
+	/// </summary>
+	/// <remarks>
+	/// This class can be used to provide an add-in management command line tool to applications.
+	/// </remarks>
 	public class SetupTool
 	{
 		Hashtable options = new Hashtable ();
@@ -51,6 +57,12 @@ namespace Mono.Addins.Setup
 		
 		int verbose = 1;
 		
+		/// <summary>
+		/// Creates a new instance
+		/// </summary>
+		/// <param name="registry">
+		/// Add-in registry to manage.
+		/// </param>
 		public SetupTool (AddinRegistry registry)
 		{
 			this.registry = registry;
@@ -58,26 +70,51 @@ namespace Mono.Addins.Setup
 			CreateCommands ();
 		}
 		
+		/// <summary>
+		/// Display name of the host application
+		/// </summary>
 		public string ApplicationName {
 			get { return applicationName; }
 			set { applicationName = value; }
 		}
 		
+		/// <summary>
+		/// Default add-in namespace of the application (optional). If set, only add-ins that belong to that namespace
+		/// will be shown in add-in lists.
+		/// </summary>
 		public string ApplicationNamespace {
 			get { return service.ApplicationNamespace; }
 			set { service.ApplicationNamespace = value; }
 		}
 		
+		/// <summary>
+		/// Enables or disables verbose output
+		/// </summary>
 		public bool VerboseOutput {
 			get { return verbose > 1; }
 			set { verbose = value ? 2 : 1; }
 		}
 		
+		/// <summary>
+		/// Sets or gets the verbose output level (0: normal output, 1:verbose, 2+:extra verbose)
+		/// </summary>
 		public int VerboseOutputLevel {
 			get { return verbose; }
 			set { verbose = value; }
 		}
 
+		/// <summary>
+		/// Runs the command line tool.
+		/// </summary>
+		/// <param name="args">
+		/// Array that contains the command line arguments
+		/// </param>
+		/// <param name="firstArgumentIndex">
+		/// Index of the arguments array that has the first argument for the management tool
+		/// </param>
+		/// <returns>
+		/// 0 if it succeeds. != 0 otherwise
+		/// </returns>
 		public int Run (string[] args, int firstArgumentIndex)
 		{
 			string[] aa = new string [args.Length - firstArgumentIndex];
@@ -85,6 +122,15 @@ namespace Mono.Addins.Setup
 			return Run (aa);
 		}
 		
+		/// <summary>
+		/// Runs the command line tool.
+		/// </summary>
+		/// <param name="args">
+		/// Command line arguments
+		/// </param>
+		/// <returns>
+		/// 0 if it succeeds. != 0 otherwise
+		/// </returns>
 		public int Run (string[] args)
 		{
 			if (args.Length == 0) {
@@ -339,22 +385,65 @@ namespace Mono.Addins.Setup
 				service.Repositories.RegisterRepository (new ConsoleProgressStatus (verbose), rep);
 		}
 		
+		string GetRepositoryUrl (string url)
+		{
+			AddinRepository[] reps = GetRepositoryList ();
+			int nr;
+			if (int.TryParse (url, out nr)) {
+				if (nr < 0 || nr >= reps.Length)
+					throw new InstallException ("Invalid repository number.");
+				return reps[nr].Url;
+			} else {
+				if (!service.Repositories.ContainsRepository (url))
+					throw new InstallException ("Repository not registered.");
+				return url;
+			}
+		}
+		
 		void RemoveRepository (string[] args)
+		{
+			foreach (string rep in args) {
+				service.Repositories.RemoveRepository (GetRepositoryUrl (rep));
+			}
+		}
+		
+		void EnableRepository (string[] args)
 		{
 			foreach (string rep in args)
-				service.Repositories.RemoveRepository (rep);
+				service.Repositories.SetRepositoryEnabled (GetRepositoryUrl(rep), true);
+		}
+		
+		void DisableRepository (string[] args)
+		{
+			foreach (string rep in args)
+				service.Repositories.SetRepositoryEnabled (GetRepositoryUrl(rep), false);
+		}
+		
+		AddinRepository[] GetRepositoryList ()
+		{
+			AddinRepository[] reps = service.Repositories.GetRepositories ();
+			Array.Sort (reps, (r1,r2) => r1.Title.CompareTo(r2.Title));
+			return reps;
 		}
 		
 		void ListRepositories (string[] args)
 		{
-			AddinRepository[] reps = service.Repositories.GetRepositories ();
+			AddinRepository[] reps = GetRepositoryList ();
 			if (reps.Length == 0) {
 				Console.WriteLine ("No repositories have been registered.");
 				return;
 			}
+			int n = 0;
 			Console.WriteLine ("Registered repositories:");
 			foreach (RepositoryRecord rep in reps) {
-				Console.WriteLine (" - " + rep.Title);
+				string num = n.ToString ();
+				Console.Write (num + ") ");
+				if (!rep.Enabled)
+					Console.Write ("(Disabled) ");
+				Console.WriteLine (rep.Title);
+				if (rep.Title != rep.Url)
+					Console.WriteLine (new string (' ', num.Length + 2) + rep.Url);
+				n++;
 			}
 		}
 		
@@ -768,6 +857,30 @@ namespace Mono.Addins.Setup
 			arguments = (string[]) list.ToArray (typeof(string));
 		}
 		
+		/// <summary>
+		/// Adds a custom command to the add-in manager
+		/// </summary>
+		/// <param name="category">
+		/// Category under which the command has to be shown in the help text
+		/// </param>
+		/// <param name="command">
+		/// Name of the command
+		/// </param>
+		/// <param name="shortName">
+		/// Short name of the command (it's an alias of the normal name)
+		/// </param>
+		/// <param name="arguments">
+		/// Formal description of the arguments that the command accepts. For example: "[addin-id|addin-file] [--xml] [--all] [--full] [--namespace <namespace>]"
+		/// </param>
+		/// <param name="description">
+		/// Short description of the command
+		/// </param>
+		/// <param name="longDescription">
+		/// Long description of the command
+		/// </param>
+		/// <param name="handler">
+		/// Delegate to be invoked to run the command
+		/// </param>
 		public void AddCommand (string category, string command, string shortName, string arguments, string description, string longDescription, SetupCommandHandler handler)
 		{
 			SetupCommand cmd = new SetupCommand (category, command, shortName, handler);
@@ -795,11 +908,19 @@ namespace Mono.Addins.Setup
 			return null;
 		}
 
+		/// <summary>
+		/// Prints help about the add-in management tool, or about a specific command
+		/// </summary>
+		/// <param name="parms">
+		/// Optional command name and arguments
+		/// </param>
 		public void PrintHelp (params string[] parms)
 		{
 			if (parms.Length == 0) {
 				string lastCat = null;
 				foreach (SetupCommand cmd in commands) {
+					if (cmd.Command == "help")
+						continue;
 					if (lastCat != cmd.Category) {
 						Console.WriteLine ();
 						Console.WriteLine (cmd.Category + ":");
@@ -823,7 +944,11 @@ namespace Mono.Addins.Setup
 					Console.WriteLine ();
 					Console.WriteLine ("Usage: {0}{1}", setupAppName, cmd.Usage);
 					Console.WriteLine ();
-					Console.WriteLine (cmd.LongDescription);
+					
+					TextFormatter fm = new TextFormatter ();
+					fm.Wrap = WrappingType.Word;
+					fm.Append (cmd.LongDescription);
+					Console.WriteLine (fm.ToString ());
 				}
 				else
 					Console.WriteLine ("Unknown command: " + parms [0]);
@@ -841,7 +966,7 @@ namespace Mono.Addins.Setup
 			cmd.Usage = "[package-name|package-file] ...";
 			cmd.AppendDesc ("Installs an add-in or set of addins. The command argument is a list");
 			cmd.AppendDesc ("of files and/or package names. If a package name is provided");
-			cmd.AppendDesc ("the package will be looked out in the registered repositories.");
+			cmd.AppendDesc ("the package will be looked up in the registered repositories.");
 			cmd.AppendDesc ("A specific add-in version can be specified by appending it to.");
 			cmd.AppendDesc ("the package name using '/' as a separator, like in this example:");
 			cmd.AppendDesc ("MonoDevelop.SourceEditor/0.9.1");
@@ -895,8 +1020,28 @@ namespace Mono.Addins.Setup
 
 			cmd = new SetupCommand (cat, "rep-remove", "rr", new SetupCommandHandler (RemoveRepository));
 			cmd.Description = "Unregisters repositories.";
-			cmd.Usage = "<url> ...";
+			cmd.Usage = "<url or number> ...";
 			cmd.AppendDesc ("Unregisters an add-in repository. Several URLs can be provided.");
+			cmd.AppendDesc ("Instead of an url, a repository number can be used (repository numbers are");
+			cmd.AppendDesc ("shown by the rep-list command.");
+			commands.Add (cmd);
+
+			cmd = new SetupCommand (cat, "rep-enable", "re", new SetupCommandHandler (EnableRepository));
+			cmd.Description = "Enables repositories.";
+			cmd.Usage = "<url or number> ...";
+			cmd.AppendDesc ("Enables an add-in repository which has been disabled. Several URLs can be");
+			cmd.AppendDesc ("provided. Instead of an url, a repository number can be used (repository");
+			cmd.AppendDesc ("numbers are shown by the rep-list command.");
+			commands.Add (cmd);
+
+			cmd = new SetupCommand (cat, "rep-disable", "rd", new SetupCommandHandler (DisableRepository));
+			cmd.Description = "Disables repositories.";
+			cmd.Usage = "<url> ...";
+			cmd.AppendDesc ("Disables an add-in repository. Several URLs can be provided");
+			cmd.AppendDesc ("When a repository is disabled, it will be ignored when using the update and");
+			cmd.AppendDesc ("install commands.");
+			cmd.AppendDesc ("Instead of an url, a repository number can be used (repository numbers are");
+			cmd.AppendDesc ("shown by the rep-list command.");
 			commands.Add (cmd);
 
 			cmd = new SetupCommand (cat, "rep-update", "ru", new SetupCommandHandler (UpdateAvailableAddins));
@@ -925,10 +1070,10 @@ namespace Mono.Addins.Setup
 			cmd = new SetupCommand (cat, "info", null, new SetupCommandHandler (PrintAddinInfo));
 			cmd.Usage = "[addin-id|addin-file] [--xml] [--all] [--full] [--namespace <namespace>]";
 			cmd.Description = "Prints information about add-ins.";
-			cmd.AppendDesc ("Prints information about add-ins. Options:");
-			cmd.AppendDesc (" --xml: Dump the information using an XML format.");
-			cmd.AppendDesc (" --all: Dump information from all add-ins.");
-			cmd.AppendDesc (" --full: Include add-ins which don't define extension points.");
+			cmd.AppendDesc ("Prints information about add-ins. Options:\n");
+			cmd.AppendDesc (" --xml: Dump the information using an XML format.\n");
+			cmd.AppendDesc (" --all: Dump information from all add-ins.\n");
+			cmd.AppendDesc (" --full: Include add-ins which don't define extension points.\n");
 			cmd.AppendDesc (" --namespace ns: Include only add-ins from the specified 'ns' namespace.");
 			commands.Add (cmd);
 
@@ -996,7 +1141,7 @@ namespace Mono.Addins.Setup
 		
 		public void AppendDesc (string s)
 		{
-			LongDescription += s + "\n";
+			LongDescription += s + " ";
 		}
 		
 		public string Category;
@@ -1022,6 +1167,9 @@ namespace Mono.Addins.Setup
 		public string LongDescription = "";
 	}
 	
+	/// <summary>
+	/// A command handler
+	/// </summary>
 	public delegate void SetupCommandHandler (string[] args);
 }
 
