@@ -36,6 +36,7 @@ using System.Xml;
 using System.Reflection;
 using Mono.Addins.Description;
 using System.Collections.Generic;
+using System.Linq;
 
 namespace Mono.Addins.Database
 {
@@ -46,9 +47,9 @@ namespace Mono.Addins.Database
 		
 		public const string VersionTag = "001";
 
-		ArrayList allSetupInfos;
-		ArrayList addinSetupInfos;
-		ArrayList rootSetupInfos;
+		List<Addin> allSetupInfos;
+		List<Addin> addinSetupInfos;
+		List<Addin> rootSetupInfos;
 		internal static bool RunningSetupProcess;
 		bool fatalDatabseError;
 		Hashtable cachedAddinSetupInfos = new Hashtable ();
@@ -60,12 +61,14 @@ namespace Mono.Addins.Database
 		AddinRegistry registry;
 		int lastDomainId;
 		AddinEngine addinEngine;
+		AddinFileSystemExtension fs = new AddinFileSystemExtension ();
+		List<object> extensions = new List<object> ();
 		
 		public AddinDatabase (AddinEngine addinEngine, AddinRegistry registry)
 		{
 			this.addinEngine = addinEngine;
 			this.registry = registry;
-			addinDbDir = Path.Combine (registry.RegistryPath, "addin-db-" + VersionTag);
+			addinDbDir = Path.Combine (registry.AddinCachePath, "addin-db-" + VersionTag);
 			fileDatabase = new FileDatabase (AddinDbDir);
 		}
 		
@@ -73,6 +76,10 @@ namespace Mono.Addins.Database
 			get { return addinDbDir; }
 		}
 		
+		public AddinFileSystemExtension FileSystem {
+			get { return fs; }
+		}
+		
 		public string AddinCachePath {
 			get { return Path.Combine (AddinDbDir, "addin-data"); }
 		}
@@ -103,6 +110,12 @@ namespace Mono.Addins.Database
 			}
 		}
 		
+		public AddinRegistry Registry {
+			get {
+				return this.registry;
+			}
+		}
+		
 		public void Clear ()
 		{
 			if (Directory.Exists (AddinCachePath))
@@ -111,6 +124,30 @@ namespace Mono.Addins.Database
 				Directory.Delete (AddinFolderCachePath, true);
 		}
 		
+		public void CopyExtensions (AddinDatabase other)
+		{
+			foreach (object o in other.extensions)
+				RegisterExtension (o);
+		}
+		
+		public void RegisterExtension (object extension)
+		{
+			extensions.Add (extension);
+			if (extension is AddinFileSystemExtension)
+				fs = (AddinFileSystemExtension) extension;
+			else
+				throw new NotSupportedException ();
+		}
+		
+		public void UnregisterExtension (object extension)
+		{
+			extensions.Remove (extension);
+			if ((extension as AddinFileSystemExtension) == fs)
+				fs = new AddinFileSystemExtension ();
+			else
+				throw new InvalidOperationException ();
+		}
+		
 		public ExtensionNodeSet FindNodeSet (string domain, string addinId, string id)
 		{
 			return FindNodeSet (domain, addinId, id, new Hashtable ());
@@ -145,53 +182,75 @@ namespace Mono.Addins.Database
 			return null;
 		}
 		
-		public ArrayList GetInstalledAddins (string domain, AddinType type)
+		public IEnumerable<Addin> GetInstalledAddins (string domain, AddinSearchFlagsInternal flags)
 		{
-			if (type == AddinType.All) {
+			if (domain == null)
+				domain = registry.CurrentDomain;
+			
+			// Get the cached list if the add-in list has already been loaded.
+			// The domain doesn't have to be checked again, since it is always the same
+			
+			IEnumerable<Addin> result = null;
+			
+			if ((flags & AddinSearchFlagsInternal.IncludeAll) == AddinSearchFlagsInternal.IncludeAll) {
 				if (allSetupInfos != null)
-					return allSetupInfos;
+					result = allSetupInfos;
 			}
-			else if (type == AddinType.Addin) {
+			else if ((flags & AddinSearchFlagsInternal.IncludeAddins) == AddinSearchFlagsInternal.IncludeAddins) {
 				if (addinSetupInfos != null)
-					return addinSetupInfos;
+					result = addinSetupInfos;
 			}
 			else {
 				if (rootSetupInfos != null)
-					return rootSetupInfos;
+					result = rootSetupInfos;
 			}
-		
-			InternalCheck (domain);
 			
-			using (fileDatabase.LockRead ()) {
-				return InternalGetInstalledAddins (domain, null, type);
+			if (result == null) {
+				InternalCheck (domain);
+				using (fileDatabase.LockRead ()) {
+					result = InternalGetInstalledAddins (domain, null, flags & ~AddinSearchFlagsInternal.LatestVersionsOnly);
+				}
 			}
+			
+			if ((flags & AddinSearchFlagsInternal.LatestVersionsOnly) == AddinSearchFlagsInternal.LatestVersionsOnly)
+				result = result.Where (a => a.IsLatestVersion);
+			
+			if ((flags & AddinSearchFlagsInternal.ExcludePendingUninstall) == AddinSearchFlagsInternal.ExcludePendingUninstall)
+				result = result.Where (a => !IsRegisteredForUninstall (a.Description.Domain, a.Id));
+			
+			return result;
 		}
 		
-		ArrayList InternalGetInstalledAddins (string domain, AddinType type)
+		IEnumerable<Addin> InternalGetInstalledAddins (string domain, AddinSearchFlagsInternal type)
 		{
 			return InternalGetInstalledAddins (domain, null, type);
 		}
 		
-		ArrayList InternalGetInstalledAddins (string domain, string idFilter, AddinType type)
+		IEnumerable<Addin> InternalGetInstalledAddins (string domain, string idFilter, AddinSearchFlagsInternal type)
 		{
+			if ((type & AddinSearchFlagsInternal.LatestVersionsOnly) != 0)
+				throw new InvalidOperationException ("LatestVersionsOnly flag not supported here");
+			
 			if (allSetupInfos == null) {
-				ArrayList alist = new ArrayList ();
+				Dictionary<string,Addin> adict = new Dictionary<string, Addin> ();
 
 				// Global add-ins are valid for any private domain
 				if (domain != AddinDatabase.GlobalDomain)
-					FindInstalledAddins (alist, AddinDatabase.GlobalDomain, idFilter);
+					FindInstalledAddins (adict, AddinDatabase.GlobalDomain, idFilter);
 
-				FindInstalledAddins (alist, domain, idFilter);
+				FindInstalledAddins (adict, domain, idFilter);
+				List<Addin> alist = new List<Addin> (adict.Values);
+				UpdateLastVersionFlags (alist);
 				if (idFilter != null)
 					return alist;
 				allSetupInfos = alist;
 			}
-			if (type == AddinType.All)
+			if ((type & AddinSearchFlagsInternal.IncludeAll) == AddinSearchFlagsInternal.IncludeAll)
 				return FilterById (allSetupInfos, idFilter);
 			
-			if (type == AddinType.Addin) {
+			if ((type & AddinSearchFlagsInternal.IncludeAddins) == AddinSearchFlagsInternal.IncludeAddins) {
 				if (addinSetupInfos == null) {
-					addinSetupInfos = new ArrayList ();
+					addinSetupInfos = new List<Addin> ();
 					foreach (Addin adn in allSetupInfos)
 						if (!adn.Description.IsRoot)
 							addinSetupInfos.Add (adn);
@@ -200,7 +259,7 @@ namespace Mono.Addins.Database
 			}
 			else {
 				if (rootSetupInfos == null) {
-					rootSetupInfos = new ArrayList ();
+					rootSetupInfos = new List<Addin> ();
 					foreach (Addin adn in allSetupInfos)
 						if (adn.Description.IsRoot)
 							rootSetupInfos.Add (adn);
@@ -209,19 +268,14 @@ namespace Mono.Addins.Database
 			}
 		}
 		
-		ArrayList FilterById (ArrayList addins, string id)
+		IEnumerable<Addin> FilterById (List<Addin> addins, string id)
 		{
 			if (id == null)
 				return addins;
-			ArrayList list = new ArrayList ();
-			foreach (Addin adn in addins) {
-				if (Addin.GetIdName (adn.Id) == id)
-					list.Add (adn);
-			}
-			return list;
+			return addins.Where (a => Addin.GetIdName (a.Id) == id);
 		}
 
-		void FindInstalledAddins (ArrayList result, string domain, string idFilter)
+		void FindInstalledAddins (Dictionary<string,Addin> result, string domain, string idFilter)
 		{
 			if (idFilter == null) 
 				idFilter = "*";
@@ -229,10 +283,31 @@ namespace Mono.Addins.Database
 			if (Directory.Exists (dir)) {
 				foreach (string file in fileDatabase.GetDirectoryFiles (dir, idFilter + ",*.maddin")) {
 					string id = Path.GetFileNameWithoutExtension (file);
-					result.Add (GetInstalledDomainAddin (domain, id, true, false, false));
+					if (!result.ContainsKey (id)) {
+						var adesc = GetInstalledDomainAddin (domain, id, true, false, false);
+						if (adesc != null)
+							result.Add (id, adesc);
+					}
 				}
 			}
 		}
+		
+		void UpdateLastVersionFlags (List<Addin> addins)
+		{
+			Dictionary<string,string> versions = new Dictionary<string, string> ();
+			foreach (Addin a in addins) {
+				string last;
+				string id, version;
+				Addin.GetIdParts (a.Id, out id, out version);
+				if (!versions.TryGetValue (id, out last) || Addin.CompareVersions (last, version) > 0)
+					versions [id] = version;
+			}
+			foreach (Addin a in addins) {
+				string id, version;
+				Addin.GetIdParts (a.Id, out id, out version);
+				a.IsLatestVersion = versions [id] == version;
+			}
+		}
 
 		public Addin GetInstalledAddin (string domain, string id)
 		{
@@ -298,7 +373,7 @@ namespace Mono.Addins.Database
 					string version, name, bestVersion = null;
 					Addin.GetIdParts (id, out name, out version);
 					
-					foreach (Addin ia in InternalGetInstalledAddins (domain, name, AddinType.All)) 
+					foreach (Addin ia in InternalGetInstalledAddins (domain, name, AddinSearchFlagsInternal.IncludeAll)) 
 					{
 						if ((!enabledOnly || ia.Enabled) &&
 						    (version.Length == 0 || ia.SupportsVersion (version)) && 
@@ -419,7 +494,7 @@ namespace Mono.Addins.Database
 			try {
 				string idName = Addin.GetIdName (id);
 				
-				foreach (Addin ainfo in GetInstalledAddins (domain, AddinType.Addin)) {
+				foreach (Addin ainfo in GetInstalledAddins (domain, AddinSearchFlagsInternal.IncludeAddins)) {
 					foreach (Dependency dep in ainfo.AddinInfo.Dependencies) {
 						AddinDependency adep = dep as AddinDependency;
 						if (adep == null)
@@ -451,8 +526,25 @@ namespace Mono.Addins.Database
 
 			if (addinEngine != null && addinEngine.IsInitialized)
 				addinEngine.UnloadAddin (id);
-		}		
+		}
+		
+		public void RegisterForUninstall (string domain, string id, IEnumerable<string> files)
+		{
+			DisableAddin (domain, id);
+			Configuration.RegisterForUninstall (id, files);
+			SaveConfiguration ();
+		}
 
+		public bool IsRegisteredForUninstall (string domain, string addinId)
+		{
+			return Configuration.IsRegisteredForUninstall (addinId);
+		}
+		
+		internal bool HasPendingUninstalls (string domain)
+		{
+			return Configuration.HasPendingUninstalls;
+		}
+		
 		internal string GetDescriptionPath (string domain, string id)
 		{
 			return Path.Combine (Path.Combine (AddinCachePath, domain), id + ".maddin");
@@ -474,7 +566,7 @@ namespace Mono.Addins.Database
 				Update (null, domain);
 		}
 		
-		void GenerateAddinExtensionMapsInternal (IProgressStatus monitor, ArrayList addinsToUpdate, ArrayList addinsToUpdateRelations, ArrayList removedAddins)
+		void GenerateAddinExtensionMapsInternal (IProgressStatus monitor, List<string> addinsToUpdate, List<string> addinsToUpdateRelations, List<string> removedAddins)
 		{
 			AddinUpdateData updateData = new AddinUpdateData (this, monitor);
 			
@@ -503,7 +595,10 @@ namespace Mono.Addins.Database
 				if (monitor.LogLevel > 2)
 					monitor.Log ("Doing a partial registry update.\nAdd-ins to be updated:");
 				// Get the files and ids of all add-ins that have to be updated
-				foreach (string sa in addinsToUpdate) {
+				// Include removed add-ins: if there are several instances of the same add-in, removing one of
+				// them will make other instances to show up. If there is a single instance, its files are
+				// already removed.
+				foreach (string sa in addinsToUpdate.Union (removedAddins)) {
 					changedAddins [sa] = sa;
 					if (monitor.LogLevel > 2)
 						monitor.Log (" - " + sa);
@@ -533,9 +628,6 @@ namespace Mono.Addins.Database
 						}
 					}
 				}
-				
-				foreach (string s in removedAddins)
-					changedAddins [s] = s;
 			}
 			else {
 				foreach (string dom in domains)
@@ -552,7 +644,7 @@ namespace Mono.Addins.Database
 				}
 
 				// If the original file does not exist, the description can be deleted
-				if (!File.Exists (conf.AddinFile)) {
+				if (!fs.FileExists (conf.AddinFile)) {
 					SafeDelete (monitor, file);
 					continue;
 				}
@@ -568,8 +660,9 @@ namespace Mono.Addins.Database
 
 			// Sort the add-ins, to make sure add-ins are processed before
 			// all their dependencies
-			ArrayList sorted = addinHash.GetSortedAddins ();
-
+			
+			var sorted = addinHash.GetSortedAddins ();
+			
 			// Register extension points and node sets
 			foreach (AddinDescription conf in sorted)
 				CollectExtensionPointData (conf, updateData);
@@ -582,7 +675,7 @@ namespace Mono.Addins.Database
 				if (changedAddins == null || changedAddins.ContainsKey (conf.AddinId)) {
 					if (monitor.LogLevel > 2)
 						monitor.Log ("- " + conf.AddinId + " (" + conf.Domain + ")");
-					CollectExtensionData (conf, updateData);
+					CollectExtensionData (monitor, addinHash, conf, updateData);
 				}
 			}
 			
@@ -601,7 +694,7 @@ namespace Mono.Addins.Database
 				monitor.Log ("  Node sets: " + updateData.RelNodeSetTypes);
 			}
 		}
-
+		
 		void ConsolidateExtensions (AddinDescription conf)
 		{
 			// Merges extensions with the same path
@@ -755,14 +848,50 @@ namespace Mono.Addins.Database
 			}
 		}
 		
-		void CollectExtensionData (AddinDescription conf, AddinUpdateData updateData)
+		void CollectExtensionData (IProgressStatus monitor, AddinIndex addinHash, AddinDescription conf, AddinUpdateData updateData)
 		{
-			foreach (ModuleDescription module in conf.AllModules) {
-				foreach (Extension ext in module.Extensions) {
-					updateData.RelExtensions++;
-					updateData.RegisterExtension (conf, module, ext);
-					AddChildExtensions (conf, module, updateData, ext.Path, ext.ExtensionNodes, false);
+			IEnumerable<string> missingDeps = addinHash.GetMissingDependencies (conf, conf.MainModule);
+			if (missingDeps.Any ()) {
+				string w = "The add-in '" + conf.AddinId + "' could not be updated because some of its dependencies are missing or not compatible:";
+				w += BuildMissingAddinsList (addinHash, conf, missingDeps);
+				monitor.ReportWarning (w);
+				return;
+			}
+			
+			CollectModuleExtensionData (conf, conf.MainModule, updateData);
+			
+			foreach (ModuleDescription module in conf.OptionalModules) {
+				missingDeps = addinHash.GetMissingDependencies (conf, module);
+				if (missingDeps.Any ()) {
+					if (monitor.LogLevel > 1) {
+						string w = "An optional module of the add-in '" + conf.AddinId + "' could not be updated because some of its dependencies are missing or not compatible:";
+						w += BuildMissingAddinsList (addinHash, conf, missingDeps);
+					}
 				}
+				else
+					CollectModuleExtensionData (conf, module, updateData);
+			}
+		}
+		
+		string BuildMissingAddinsList (AddinIndex addinHash, AddinDescription conf, IEnumerable<string> missingDeps)
+		{
+			string w = "";
+			foreach (string dep in missingDeps) {
+				var found = addinHash.GetSimilarExistingAddin (conf, dep);
+				if (found == null)
+					w += "\n  missing: " + dep;
+				else
+					w += "\n  required: " + dep + ", found: " + found.AddinId;
+			}
+			return w;
+		}
+		
+		void CollectModuleExtensionData (AddinDescription conf, ModuleDescription module, AddinUpdateData updateData)
+		{
+			foreach (Extension ext in module.Extensions) {
+				updateData.RelExtensions++;
+				updateData.RegisterExtension (conf, module, ext);
+				AddChildExtensions (conf, module, updateData, ext.Path, ext.ExtensionNodes, false);
 			}
 		}
 		
@@ -887,6 +1016,8 @@ namespace Mono.Addins.Database
 			
 			DateTime tim = DateTime.Now;
 			
+			RunPendingUninstalls (monitor);
+			
 			Hashtable installed = new Hashtable ();
 			bool changesFound = CheckFolders (monitor, domain);
 			
@@ -902,8 +1033,7 @@ namespace Mono.Addins.Database
 				
 				if (domain != null) {
 					using (fileDatabase.LockRead ()) {
-						// Don't use AddinType.Addin here because it is too expensive
-						foreach (Addin ainfo in InternalGetInstalledAddins (domain, AddinType.All)) {
+						foreach (Addin ainfo in InternalGetInstalledAddins (domain, AddinSearchFlagsInternal.IncludeAddins)) {
 							installed [ainfo.Id] = ainfo.Id;
 						}
 					}
@@ -922,30 +1052,74 @@ namespace Mono.Addins.Database
 			// Update the currently loaded add-ins
 			if (changesFound && domain != null && addinEngine != null && addinEngine.IsInitialized) {
 				Hashtable newInstalled = new Hashtable ();
-				foreach (Addin ainfo in GetInstalledAddins (domain, AddinType.All)) {
+				foreach (Addin ainfo in GetInstalledAddins (domain, AddinSearchFlagsInternal.IncludeAddins)) {
 					newInstalled [ainfo.Id] = ainfo.Id;
 				}
 				
 				foreach (string aid in installed.Keys) {
-					if (!newInstalled.Contains (aid)) {
-						if (addinEngine.IsAddinLoaded (aid)) {
-							RuntimeAddin ra = addinEngine.GetAddin (aid);
-							if (!ra.Addin.Description.IsRoot)
-								addinEngine.UnloadAddin (aid);
-						}
-					}
+					// Always try to unload, event if the add-in was not currently loaded.
+					// Required since the add-ins has to be marked as 'disabled', to avoid
+					// extensions from this add-in to be loaded
+					if (!newInstalled.Contains (aid))
+						addinEngine.UnloadAddin (aid);
 				}
 				
 				foreach (string aid in newInstalled.Keys) {
 					if (!installed.Contains (aid)) {
 						Addin addin = addinEngine.Registry.GetAddin (aid);
-						if (addin != null && !addin.Description.IsRoot)
+						if (addin != null)
 							addinEngine.ActivateAddin (aid);
 					}
 				}
 			}
 		}
 		
+		void RunPendingUninstalls (IProgressStatus monitor)
+		{
+			bool changesDone = false;
+			
+			foreach (var adn in Configuration.GetPendingUninstalls ()) {
+				HashSet<string> files = new HashSet<string> (adn.Files);
+				if (AddinManager.CheckAssembliesLoaded (files))
+					continue;
+				
+				if (monitor.LogLevel > 1)
+					monitor.Log ("Uninstalling " + adn.AddinId);
+				
+				// Make sure all files can be deleted before doing so
+				bool canUninstall = true;
+				foreach (string f in adn.Files) {
+					if (!File.Exists (f))
+						continue;
+					try {
+						File.OpenWrite (f).Close ();
+					} catch {
+						canUninstall = false;
+						break;
+					}
+				}
+				
+				if (!canUninstall)
+					continue;
+				
+				foreach (string f in adn.Files) {
+					try {
+						if (File.Exists (f))
+							File.Delete (f);
+					} catch {
+						canUninstall = false;
+					}
+				}
+				
+				if (canUninstall) {
+					Configuration.UnregisterForUninstall (adn.AddinId);
+					changesDone = true;
+				}
+			}
+			if (changesDone)
+				SaveConfiguration ();
+		}
+		
 		void RunScannerProcess (IProgressStatus monitor)
 		{
 			ISetupHandler setup = GetSetupHandler ();
@@ -958,7 +1132,7 @@ namespace Mono.Addins.Database
 				try {
 					if (monitor.LogLevel > 1)
 						monitor.Log ("Looking for addins");
-					setup.Scan (scanMonitor, registry.RegistryPath, registry.StartupDirectory, null, (string[]) pparams.ToArray (typeof(string)));
+					setup.Scan (scanMonitor, registry, null, (string[]) pparams.ToArray (typeof(string)));
 					retry = false;
 				}
 				catch (Exception ex) {
@@ -1101,6 +1275,16 @@ namespace Mono.Addins.Database
 		}
 		
 		void InternalScanFolders (IProgressStatus monitor, AddinScanResult scanResult)
+		{
+			try {
+				fs.ScanStarted ();
+				InternalScanFolders2 (monitor, scanResult);
+			} finally {
+				fs.ScanFinished ();
+			}
+		}
+		
+		void InternalScanFolders2 (IProgressStatus monitor, AddinScanResult scanResult)
 		{
 			DateTime tim = DateTime.Now;
 			
@@ -1128,7 +1312,7 @@ namespace Mono.Addins.Database
 				AddinScanFolderInfo folderInfo;
 				bool res = ReadFolderInfo (monitor, file, out folderInfo);
 				bool validForDomain = scanResult.Domain == null || folderInfo.Domain == GlobalDomain || folderInfo.Domain == scanResult.Domain;
-				if (!res || (validForDomain && !Directory.Exists (folderInfo.Folder))) {
+				if (!res || (validForDomain && !fs.DirectoryExists (folderInfo.Folder))) {
 					if (res) {
 						// Folder has been deleted. Remove the add-ins it had.
 						scanner.UpdateDeletedAddins (monitor, folderInfo, scanResult);
@@ -1166,7 +1350,7 @@ namespace Mono.Addins.Database
 				scanner.ScanFolderRec (monitor, dir, GlobalDomain, scanResult);
 			}
 			
-			if (scanResult.CheckOnly)
+			if (scanResult.CheckOnly || !scanResult.ChangesFound)
 				return;
 			
 			// Scan the files which have been modified
@@ -1219,7 +1403,7 @@ namespace Mono.Addins.Database
 		{
 			if (!inProcess) {
 				ISetupHandler setup = GetSetupHandler ();
-				setup.GetAddinDescription (progressStatus, registry.RegistryPath, registry.StartupDirectory, Path.GetFullPath (file), outFile);
+				setup.GetAddinDescription (progressStatus, registry, Path.GetFullPath (file), outFile);
 				return;
 			}
 			
@@ -1231,7 +1415,7 @@ namespace Mono.Addins.Database
 					AddinFileInfo afi = finfo.GetAddinFileInfo (file);
 					if (afi != null && afi.IsAddin) {
 						AddinDescription adesc;
-						GetAddinDescription (progressStatus, afi.Domain, afi.AddinId, out adesc);
+						GetAddinDescription (progressStatus, afi.Domain, afi.AddinId, file, out adesc);
 						if (adesc != null)
 							adesc.Save (outFile);
 						return;
@@ -1277,7 +1461,8 @@ namespace Mono.Addins.Database
 			if (file != null)
 				return Util.LoadAssemblyForReflection (file);
 			else {
-				Console.WriteLine ("Assembly not found: " + args.Name);
+				if (!args.Name.StartsWith ("Mono.Addins.CecilReflector"))
+					Console.WriteLine ("Assembly not found: " + args.Name);
 				return null;
 			}
 		}
@@ -1294,36 +1479,104 @@ namespace Mono.Addins.Database
 			return Path.Combine (AddinFolderCachePath, s + ".data");
 		}
 		
-		internal void UninstallAddin (IProgressStatus monitor, string domain, string addinId, AddinScanResult scanResult)
+		internal void UninstallAddin (IProgressStatus monitor, string domain, string addinId, string addinFile, AddinScanResult scanResult)
 		{
-			scanResult.AddRemovedAddin (addinId);
-			string file = GetDescriptionPath (domain, addinId);
-			if (!fileDatabase.Exists (file)) {
+			AddinDescription desc;
+			
+			if (!GetAddinDescription (monitor, domain, addinId, addinFile, out desc)) {
+				// If we can't get information about the old assembly, just regenerate all relation data
+				scanResult.RegenerateRelationData = true;
 				return;
 			}
 			
-			// Add-in already existed. The dependencies of the old add-in need to be re-analized
+			scanResult.AddRemovedAddin (addinId);
+			
+			// If the add-in didn't exist, there is nothing left to do
+			
+			if (desc == null)
+				return;
+			
+			// If the add-in already existed, the dependencies of the old add-in need to be re-analized
+			
+			Util.AddDependencies (desc, scanResult);
+			if (desc.IsRoot)
+				scanResult.HostIndex.RemoveHostData (desc.AddinId, desc.AddinFile);
 
-			AddinDescription desc;
-			if (ReadAddinDescription (monitor, file, out desc)) {
-				Util.AddDependencies (desc, scanResult);
-				if (desc.IsRoot)
-					scanResult.HostIndex.RemoveHostData (desc.AddinId, desc.AddinFile);
-			} else
-				// If we can't get information about the old assembly, just regenerate all relation data
-				scanResult.RegenerateRelationData = true;
+			RemoveAddinDescriptionFile (monitor, desc.FileName);
+		}
+		
+		public bool GetAddinDescription (IProgressStatus monitor, string domain, string addinId, string addinFile, out AddinDescription description)
+		{
+			// If the same add-in is installed in different folders (in the same domain) there will be several .maddin files for it,
+			// using the suffix "_X" where X is a number > 1 (for example: someAddin,1.0.maddin, someAddin,1.0.maddin_2, someAddin,1.0.maddin_3, ...)
+			// We need to return the .maddin whose AddinFile matches the one being requested
+			
+			addinFile = Path.GetFullPath (addinFile);
+			int altNum = 1;
+			string baseFile = GetDescriptionPath (domain, addinId);
+			string file = baseFile;
+			bool failed = false;
+			
+			do {
+				if (!ReadAddinDescription (monitor, file, out description)) {
+					// Remove the AddinDescription here since it is corrupted.
+					// Avoids creating alternate versions of corrupted files when later calling SaveDescription.
+					RemoveAddinDescriptionFile (monitor, file);
+					failed = true;
+					continue;
+				}
+				if (description == null)
+					break;
+				if (Path.GetFullPath (description.AddinFile) == addinFile)
+					return true;
+				file = baseFile + "_" + (++altNum);
+			}
+			while (fileDatabase.Exists (file));
+			
+			// File not found. Return false only if there has been any read error.
+			description = null;
+			return failed;
+		}
+		
+		bool RemoveAddinDescriptionFile (IProgressStatus monitor, string file)
+		{
+			// Removes an add-in description and shifts up alternate instances of the description file
+			// (so xxx,1.0.maddin_2 will become xxx,1.0.maddin, xxx,1.0.maddin_3 -> xxx,1.0.maddin_2, etc)
+			
+			if (!SafeDelete (monitor, file))
+				return false;
+			
+			int dversion;
+			if (file.EndsWith (".maddin"))
+				dversion = 2;
+			else {
+				int i = file.LastIndexOf ('_');
+				dversion = 1 + int.Parse (file.Substring (i + 1));
+				file = file.Substring (0, i);
+			}
 
-			SafeDelete (monitor, file);
+			while (fileDatabase.Exists (file + "_" + dversion)) {
+				string newFile = dversion == 2 ? file : file + "_" + (dversion-1);
+				try {
+					fileDatabase.Rename (file + "_" + dversion, newFile);
+				} catch (Exception ex) {
+					if (monitor.LogLevel > 1) {
+						monitor.Log ("Could not rename file '" + file + "_" + dversion + "' to '" + newFile + "'");
+						monitor.Log (ex.ToString ());
+					}
+				}
+				dversion++;
+			}
 			string dir = Path.GetDirectoryName (file);
 			if (fileDatabase.DirectoryIsEmpty (dir))
 				SafeDeleteDir (monitor, dir);
-			SafeDeleteDir (monitor, Path.Combine (AddinPrivateDataPath, Path.GetFileNameWithoutExtension (file)));
-		}
-		
-		public bool GetAddinDescription (IProgressStatus monitor, string domain, string addinId, out AddinDescription description)
-		{
-			string file = GetDescriptionPath (domain, addinId);
-			return ReadAddinDescription (monitor, file, out description);
+			
+			if (dversion == 2) {
+				// All versions of the add-in removed.
+				SafeDeleteDir (monitor, Path.Combine (AddinPrivateDataPath, Path.GetFileNameWithoutExtension (file)));
+			}
+			
+			return true;
 		}
 		
 		public bool ReadAddinDescription (IProgressStatus monitor, string file, out AddinDescription description)
@@ -1353,6 +1606,14 @@ namespace Mono.Addins.Database
 					string dir = Path.GetDirectoryName (file);
 					if (!fileDatabase.DirExists (dir))
 						fileDatabase.CreateDir (dir);
+					if (fileDatabase.Exists (file)) {
+						// Another AddinDescription already exists with the same name.
+						// Create an alternate AddinDescription file
+						int altNum = 2;
+						while (fileDatabase.Exists (file + "_" + altNum))
+							altNum++;
+						file = file + "_" + altNum;
+					}
 					desc.SaveBinary (fileDatabase, file);
 				}
 				return true;
@@ -1500,13 +1761,18 @@ namespace Mono.Addins.Database
 //			if (Util.IsMono)
 //				return new SetupProcess ();
 //			else
+			if (fs.RequiresIsolation)
 				return new SetupDomain ();
+			else
+				return new SetupLocal ();
 		}
 		
 		public void ResetConfiguration ()
 		{
 			if (File.Exists (ConfigFile))
 				File.Delete (ConfigFile);
+			config = null;
+			ResetCachedData ();
 		}
 		
 		DatabaseConfiguration Configuration {
@@ -1569,29 +1835,27 @@ namespace Mono.Addins.Database
 	
 	class AddinIndex
 	{
-		Hashtable addins = new Hashtable ();
+		Dictionary<string, List<AddinDescription>> addins = new Dictionary<string, List<AddinDescription>> ();
 		
 		public void Add (AddinDescription desc)
 		{
 			string id = Addin.GetFullId (desc.Namespace, desc.LocalId, null);
-			ArrayList list = (ArrayList) addins [id];
-			if (list == null) {
-				list = new ArrayList (); 
-				addins [id] = list;
-			}
+			List<AddinDescription> list;
+			if (!addins.TryGetValue (id, out list))
+				addins [id] = list = new List<AddinDescription> ();
 			list.Add (desc);
 		}
 		
-		ArrayList FindDescriptions (string domain, string fullid)
+		List<AddinDescription> FindDescriptions (string domain, string fullid)
 		{
 			// Returns all registered add-ins which are compatible with the provided
 			// fullid. Compatible means that the id is the same and the version is within
 			// the range of compatible versions of the add-in.
 			
-			ArrayList res = new ArrayList ();
+			var res = new List<AddinDescription> ();
 			string id = Addin.GetIdName (fullid);
-			ArrayList list = (ArrayList) addins [id];
-			if (list == null)
+			List<AddinDescription> list;
+			if (!addins.TryGetValue (id, out list))
 				return res;
 			string version = Addin.GetIdVersion (fullid);
 			foreach (AddinDescription desc in list) {
@@ -1601,12 +1865,38 @@ namespace Mono.Addins.Database
 			return res;
 		}
 		
-		public ArrayList GetSortedAddins ()
+		public IEnumerable<string> GetMissingDependencies (AddinDescription desc, ModuleDescription mod)
 		{
-			Hashtable inserted = new Hashtable ();
-			Hashtable lists = new Hashtable ();
+			foreach (Dependency dep in mod.Dependencies) {
+				AddinDependency adep = dep as AddinDependency;
+				if (adep == null)
+					continue;
+				var descs = FindDescriptions (desc.Domain, adep.FullAddinId);
+				if (descs.Count == 0)
+					yield return adep.FullAddinId;
+			}
+		}
+		
+		public AddinDescription GetSimilarExistingAddin (AddinDescription conf, string addinId)
+		{
+			string domain = conf.Domain;
+			List<AddinDescription> list;
+			if (!addins.TryGetValue (Addin.GetIdName (addinId), out list))
+				return null;
+			string version = Addin.GetIdVersion (addinId);
+			foreach (AddinDescription desc in list) {
+				if ((desc.Domain == domain || domain == AddinDatabase.GlobalDomain) && !desc.SupportsVersion (version))
+					return desc;
+			}
+			return null;
+		}
+		
+		public List<AddinDescription> GetSortedAddins ()
+		{
+			var inserted = new HashSet<string> ();
+			var lists = new Dictionary<string,List<AddinDescription>> ();
 			
-			foreach (ArrayList dlist in addins.Values) {
+			foreach (List<AddinDescription> dlist in addins.Values) {
 				foreach (AddinDescription desc in dlist)
 					InsertSortedAddin (inserted, lists, desc);
 			}
@@ -1614,53 +1904,53 @@ namespace Mono.Addins.Database
 			// Merge all domain lists into a single list.
 			// Make sure the global domain is inserted the last
 			
-			ArrayList global = (ArrayList) lists [AddinDatabase.GlobalDomain];
+			List<AddinDescription> global;
+			lists.TryGetValue (AddinDatabase.GlobalDomain, out global);
 			lists.Remove (AddinDatabase.GlobalDomain);
 			
-			ArrayList list = new ArrayList ();
-			foreach (ArrayList dl in lists.Values) {
-				list.AddRange (dl);
+			List<AddinDescription> sortedAddins = new List<AddinDescription> ();
+			foreach (var dl in lists.Values) {
+				sortedAddins.AddRange (dl);
 			}
 			if (global != null)
-				list.AddRange (global);
-			return list;
+				sortedAddins.AddRange (global);
+			return sortedAddins;
 		}
 
-		void InsertSortedAddin (Hashtable inserted, Hashtable lists, AddinDescription desc)
+		void InsertSortedAddin (HashSet<string> inserted, Dictionary<string,List<AddinDescription>> lists, AddinDescription desc)
 		{
 			string sid = desc.AddinId + " " + desc.Domain;
-			if (inserted.ContainsKey (sid))
+			if (!inserted.Add (sid))
 				return;
-			inserted [sid] = desc;
+
 			foreach (ModuleDescription mod in desc.AllModules) {
 				foreach (Dependency dep in mod.Dependencies) {
 					AddinDependency adep = dep as AddinDependency;
 					if (adep == null)
 						continue;
-					ArrayList descs = FindDescriptions (desc.Domain, adep.FullAddinId);
+					var descs = FindDescriptions (desc.Domain, adep.FullAddinId);
 					if (descs.Count > 0) {
 						foreach (AddinDescription sd in descs)
 							InsertSortedAddin (inserted, lists, sd);
 					}
-//					else 
-//						Console.WriteLine ("NOT FOUND: " + adep.FullAddinId + " " + desc.Domain + " from " + sid);
 				}
 			}
-			ArrayList list = (ArrayList) lists [desc.Domain];
-			if (list == null) {
-				list = new ArrayList ();
-				lists [desc.Domain] = list;
-			}
+			List<AddinDescription> list;
+			if (!lists.TryGetValue (desc.Domain, out list))
+				lists [desc.Domain] = list = new List<AddinDescription> ();
 			
 			list.Add (desc);
 		}
 	}
 	
-	enum AddinType
+	// Keep in sync with AddinSearchFlags
+	enum AddinSearchFlagsInternal
 	{
-		Addin,
-		Root,
-		All
+		IncludeAddins = 1,
+		IncludeRoots = 1 << 1,
+		IncludeAll = IncludeAddins | IncludeRoots,
+		LatestVersionsOnly = 1 << 3,
+		ExcludePendingUninstall = 1 << 4
 	}
 }
 
